package org.apache.bcel.verifier.statics;

import org.apache.bcel.generic.TABLESWITCH;
import org.apache.bcel.generic.PUTSTATIC;
import org.apache.bcel.generic.NEWARRAY;
import org.apache.bcel.generic.NEW;
import org.apache.bcel.generic.MULTIANEWARRAY;
import org.apache.bcel.generic.LSTORE;
import org.apache.bcel.generic.LOOKUPSWITCH;
import org.apache.bcel.generic.LoadClass;
import org.apache.bcel.generic.LLOAD;
import org.apache.bcel.classfile.ConstantDouble;
import org.apache.bcel.classfile.ConstantLong;
import org.apache.bcel.generic.LDC2_W;
import org.apache.bcel.classfile.ConstantString;
import org.apache.bcel.classfile.ConstantFloat;
import org.apache.bcel.classfile.ConstantInteger;
import org.apache.bcel.generic.LDC;
import org.apache.bcel.generic.ISTORE;
import org.apache.bcel.generic.INVOKEINTERFACE;
import org.apache.bcel.verifier.VerifierFactory;
import org.apache.bcel.classfile.ConstantInterfaceMethodref;
import org.apache.bcel.classfile.ConstantUtf8;
import org.apache.bcel.classfile.ConstantNameAndType;
import org.apache.bcel.classfile.ConstantMethodref;
import org.apache.bcel.generic.INVOKESTATIC;
import org.apache.bcel.generic.INVOKESPECIAL;
import org.apache.bcel.generic.INVOKEVIRTUAL;
import org.apache.bcel.generic.INVOKEDYNAMIC;
import org.apache.bcel.generic.INSTANCEOF;
import org.apache.bcel.generic.ILOAD;
import org.apache.bcel.generic.IINC;
import org.apache.bcel.generic.GETSTATIC;
import org.apache.bcel.generic.FSTORE;
import org.apache.bcel.generic.FLOAD;
import org.apache.bcel.classfile.Field;
import org.apache.bcel.classfile.ConstantFieldref;
import org.apache.bcel.generic.DSTORE;
import org.apache.bcel.generic.DLOAD;
import org.apache.bcel.generic.CHECKCAST;
import org.apache.bcel.classfile.Constant;
import org.apache.bcel.generic.ArrayType;
import org.apache.bcel.classfile.ConstantClass;
import org.apache.bcel.generic.ANEWARRAY;
import org.apache.bcel.generic.ALOAD;
import org.apache.bcel.generic.ReferenceType;
import org.apache.bcel.generic.ObjectType;
import org.apache.bcel.generic.FieldInstruction;
import java.util.Arrays;
import org.apache.bcel.generic.Type;
import org.apache.bcel.generic.InvokeInstruction;
import org.apache.bcel.generic.EmptyVisitor;
import org.apache.bcel.classfile.LocalVariable;
import org.apache.bcel.generic.Visitor;
import org.apache.bcel.generic.ASTORE;
import org.apache.bcel.verifier.exc.StaticCodeInstructionOperandConstraintException;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.ConstantPoolGen;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.ATHROW;
import org.apache.bcel.generic.GotoInstruction;
import org.apache.bcel.generic.RET;
import org.apache.bcel.generic.ReturnInstruction;
import org.apache.bcel.generic.BREAKPOINT;
import org.apache.bcel.generic.IMPDEP2;
import org.apache.bcel.generic.IMPDEP1;
import org.apache.bcel.verifier.exc.StaticCodeInstructionConstraintException;
import org.apache.bcel.classfile.Method;
import org.apache.bcel.classfile.JavaClass;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.StaticCodeConstraintException;
import org.apache.bcel.classfile.ClassFormatException;
import org.apache.bcel.verifier.exc.InvalidMethodException;
import org.apache.bcel.Repository;
import org.apache.bcel.verifier.VerificationResult;
import org.apache.bcel.classfile.CodeException;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.LineNumber;
import org.apache.bcel.classfile.LineNumberTable;
import org.apache.bcel.classfile.LocalVariableTable;
import org.apache.bcel.verifier.exc.ClassConstraintException;
import org.apache.bcel.classfile.Code;
import org.apache.bcel.generic.InstructionList;
import org.apache.bcel.verifier.Verifier;
import org.apache.bcel.verifier.PassVerifier;

public final class Pass3aVerifier extends PassVerifier
{
    private final Verifier myOwner;
    private final int methodNo;
    private InstructionList instructionList;
    private Code code;
    
    private static boolean contains(final int[] ints, final int i) {
        for (final int k : ints) {
            if (k == i) {
                return true;
            }
        }
        return false;
    }
    
    public Pass3aVerifier(final Verifier owner, final int methodNo) {
        this.myOwner = owner;
        this.methodNo = methodNo;
    }
    
    private void delayedPass2Checks() {
        final int[] instructionPositions = this.instructionList.getInstructionPositions();
        final int codeLength = this.code.getCode().length;
        final LineNumberTable lnt = this.code.getLineNumberTable();
        if (lnt != null) {
            final LineNumber[] lineNumbers = lnt.getLineNumberTable();
            final IntList offsets = new IntList();
            final LineNumber[] array = lineNumbers;
            final int length = array.length;
            int i = 0;
        Label_0056:
            while (i < length) {
                final LineNumber lineNumber = array[i];
                for (final int instructionPosition : instructionPositions) {
                    final int offset = lineNumber.getStartPC();
                    if (instructionPosition == offset) {
                        if (offsets.contains(offset)) {
                            this.addMessage("LineNumberTable attribute '" + this.code.getLineNumberTable() + "' refers to the same code offset ('" + offset + "') more than once which is violating the semantics [but is sometimes produced by IBM's 'jikes' compiler].");
                        }
                        else {
                            offsets.add(offset);
                        }
                        ++i;
                        continue Label_0056;
                    }
                }
                throw new ClassConstraintException("Code attribute '" + this.tostring(this.code) + "' has a LineNumberTable attribute '" + this.code.getLineNumberTable() + "' referring to a code offset ('" + lineNumber.getStartPC() + "') that does not exist.");
            }
        }
        final Attribute[] attributes;
        final Attribute[] atts = attributes = this.code.getAttributes();
        for (final Attribute att : attributes) {
            if (att instanceof LocalVariableTable) {
                ((LocalVariableTable)att).forEach(localVariable -> {
                    final int startpc2 = localVariable.getStartPC();
                    final int length5 = localVariable.getLength();
                    if (!contains(instructionPositions, startpc2)) {
                        new ClassConstraintException("Code attribute '" + this.tostring(this.code) + "' has a LocalVariableTable attribute '" + this.code.getLocalVariableTable() + "' referring to a code offset ('" + startpc2 + "') that does not exist.");
                        throw;
                    }
                    else if (!contains(instructionPositions, startpc2 + length5) && startpc2 + length5 != codeLength) {
                        new ClassConstraintException("Code attribute '" + this.tostring(this.code) + "' has a LocalVariableTable attribute '" + this.code.getLocalVariableTable() + "' referring to a code offset start_pc+length ('" + (startpc2 + length5) + "') that does not exist.");
                        throw;
                    }
                    else {
                        return;
                    }
                });
            }
        }
        final CodeException[] exceptionTable2;
        final CodeException[] exceptionTable = exceptionTable2 = this.code.getExceptionTable();
        for (final CodeException element : exceptionTable2) {
            final int startpc = element.getStartPC();
            final int endpc = element.getEndPC();
            final int handlerpc = element.getHandlerPC();
            if (startpc >= endpc) {
                throw new ClassConstraintException("Code attribute '" + this.tostring(this.code) + "' has an exception_table entry '" + element + "' that has its start_pc ('" + startpc + "') not smaller than its end_pc ('" + endpc + "').");
            }
            if (!contains(instructionPositions, startpc)) {
                throw new ClassConstraintException("Code attribute '" + this.tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its start_pc ('" + startpc + "').");
            }
            if (!contains(instructionPositions, endpc) && endpc != codeLength) {
                throw new ClassConstraintException("Code attribute '" + this.tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its end_pc ('" + startpc + "') [that is also not equal to code_length ('" + codeLength + "')].");
            }
            if (!contains(instructionPositions, handlerpc)) {
                throw new ClassConstraintException("Code attribute '" + this.tostring(this.code) + "' has an exception_table entry '" + element + "' that has a non-existant bytecode offset as its handler_pc ('" + handlerpc + "').");
            }
        }
    }
    
    @Override
    public VerificationResult do_verify() {
        try {
            if (!this.myOwner.doPass2().equals(VerificationResult.VR_OK)) {
                return VerificationResult.VR_NOTYET;
            }
            final JavaClass jc = Repository.lookupClass(this.myOwner.getClassName());
            final Method[] methods = jc.getMethods();
            if (this.methodNo >= methods.length) {
                throw new InvalidMethodException("METHOD DOES NOT EXIST!");
            }
            final Method method = methods[this.methodNo];
            this.code = method.getCode();
            if (method.isAbstract() || method.isNative()) {
                return VerificationResult.VR_OK;
            }
            try {
                this.instructionList = new InstructionList(method.getCode().getCode());
            }
            catch (final RuntimeException re) {
                return new VerificationResult(2, "Bad bytecode in the code array of the Code attribute of method '" + this.tostring(method) + "'.");
            }
            this.instructionList.setPositions(true);
            VerificationResult vr = VerificationResult.VR_OK;
            try {
                this.delayedPass2Checks();
            }
            catch (final ClassConstraintException | ClassFormatException cce) {
                vr = new VerificationResult(2, cce.getMessage());
                return vr;
            }
            try {
                this.pass3StaticInstructionChecks();
                this.pass3StaticInstructionOperandsChecks();
            }
            catch (final StaticCodeConstraintException | ClassFormatException scce) {
                vr = new VerificationResult(2, scce.getMessage());
            }
            catch (final ClassCastException cce2) {
                vr = new VerificationResult(2, "Class Cast Exception: " + cce2.getMessage());
            }
            return vr;
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    public int getMethodNo() {
        return this.methodNo;
    }
    
    private void pass3StaticInstructionChecks() {
        if (this.code.getCode().length >= 65536) {
            throw new StaticCodeInstructionConstraintException("Code array in code attribute '" + this.tostring(this.code) + "' too big: must be smaller than " + 65536 + "65536 bytes.");
        }
        for (InstructionHandle ih = this.instructionList.getStart(); ih != null; ih = ih.getNext()) {
            final Instruction i = ih.getInstruction();
            if (i instanceof IMPDEP1) {
                throw new StaticCodeInstructionConstraintException("IMPDEP1 must not be in the code, it is an illegal instruction for _internal_ JVM use!");
            }
            if (i instanceof IMPDEP2) {
                throw new StaticCodeInstructionConstraintException("IMPDEP2 must not be in the code, it is an illegal instruction for _internal_ JVM use!");
            }
            if (i instanceof BREAKPOINT) {
                throw new StaticCodeInstructionConstraintException("BREAKPOINT must not be in the code, it is an illegal instruction for _internal_ JVM use!");
            }
        }
        final Instruction last = this.instructionList.getEnd().getInstruction();
        if (!(last instanceof ReturnInstruction) && !(last instanceof RET) && !(last instanceof GotoInstruction) && !(last instanceof ATHROW)) {
            throw new StaticCodeInstructionConstraintException("Execution must not fall off the bottom of the code array. This constraint is enforced statically as some existing verifiers do - so it may be a false alarm if the last instruction is not reachable.");
        }
    }
    
    private void pass3StaticInstructionOperandsChecks() {
        try {
            final ConstantPoolGen cpg = new ConstantPoolGen(Repository.lookupClass(this.myOwner.getClassName()).getConstantPool());
            final InstOperandConstraintVisitor v = new InstOperandConstraintVisitor(cpg);
            for (InstructionHandle ih = this.instructionList.getStart(); ih != null; ih = ih.getNext()) {
                final Instruction i = ih.getInstruction();
                if (i instanceof JsrInstruction) {
                    final InstructionHandle target = ((JsrInstruction)i).getTarget();
                    if (target == this.instructionList.getStart()) {
                        throw new StaticCodeInstructionOperandConstraintException("Due to JustIce's clear definition of subroutines, no JSR or JSR_W may have a top-level instruction (such as the very first instruction, which is targeted by instruction '" + this.tostring(ih) + "' as its target.");
                    }
                    if (!(target.getInstruction() instanceof ASTORE)) {
                        throw new StaticCodeInstructionOperandConstraintException("Due to JustIce's clear definition of subroutines, no JSR or JSR_W may target anything else than an ASTORE instruction. Instruction '" + this.tostring(ih) + "' targets '" + this.tostring(target) + "'.");
                    }
                }
                ih.accept(v);
            }
        }
        catch (final ClassNotFoundException e) {
            throw new AssertionViolatedException("Missing class: " + e, e);
        }
    }
    
    protected String tostring(final Object obj) {
        String ret;
        try {
            ret = obj.toString();
        }
        catch (final RuntimeException e) {
            String s = obj.getClass().getName();
            s = s.substring(s.lastIndexOf(".") + 1);
            ret = "<<" + s + ">>";
        }
        return ret;
    }
    
    private class InstOperandConstraintVisitor extends EmptyVisitor
    {
        private final ConstantPoolGen constantPoolGen;
        
        InstOperandConstraintVisitor(final ConstantPoolGen constantPoolGen) {
            this.constantPoolGen = constantPoolGen;
        }
        
        private void constraintViolated(final Instruction i, final String message) {
            throw new StaticCodeInstructionOperandConstraintException("Instruction " + Pass3aVerifier.this.tostring(i) + " constraint violated: " + message);
        }
        
        private Method getMethod(final JavaClass jc, final InvokeInstruction invoke) {
            final Method[] methods;
            final Method[] ms = methods = jc.getMethods();
            for (final Method element : methods) {
                if (element.getName().equals(invoke.getMethodName(this.constantPoolGen)) && Type.getReturnType(element.getSignature()).equals(invoke.getReturnType(this.constantPoolGen)) && Arrays.equals(Type.getArgumentTypes(element.getSignature()), invoke.getArgumentTypes(this.constantPoolGen))) {
                    return element;
                }
            }
            return null;
        }
        
        private Method getMethodRecursive(final JavaClass jc, final InvokeInstruction invoke) throws ClassNotFoundException {
            Method m = this.getMethod(jc, invoke);
            if (m != null) {
                return m;
            }
            for (final JavaClass superclass : jc.getSuperClasses()) {
                m = this.getMethod(superclass, invoke);
                if (m != null) {
                    return m;
                }
            }
            for (final JavaClass superclass : jc.getInterfaces()) {
                m = this.getMethod(superclass, invoke);
                if (m != null) {
                    return m;
                }
            }
            return null;
        }
        
        private ObjectType getObjectType(final FieldInstruction o) {
            final ReferenceType rt = o.getReferenceType(this.constantPoolGen);
            if (rt instanceof ObjectType) {
                return (ObjectType)rt;
            }
            this.constraintViolated(o, "expecting ObjectType but got " + rt);
            return null;
        }
        
        private void indexValid(final Instruction i, final int idx) {
            if (idx < 0 || idx >= this.constantPoolGen.getSize()) {
                this.constraintViolated(i, "Illegal constant pool index '" + idx + "'.");
            }
        }
        
        private int maxLocals() {
            try {
                return Repository.lookupClass(Pass3aVerifier.this.myOwner.getClassName()).getMethods()[Pass3aVerifier.this.methodNo].getCode().getMaxLocals();
            }
            catch (final ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }
        
        @Override
        public void visitALOAD(final ALOAD o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitANEWARRAY(final ANEWARRAY o) {
            this.indexValid(o, o.getIndex());
            final Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                this.constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
            final Type t = o.getType(this.constantPoolGen);
            if (t instanceof ArrayType) {
                final int dimensions = ((ArrayType)t).getDimensions();
                if (dimensions > 255) {
                    this.constraintViolated(o, "Not allowed to create an array with more than 255 dimensions; actual: " + dimensions);
                }
            }
        }
        
        @Override
        public void visitASTORE(final ASTORE o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitCHECKCAST(final CHECKCAST o) {
            this.indexValid(o, o.getIndex());
            final Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                this.constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
        }
        
        @Override
        public void visitDLOAD(final DLOAD o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
            }
            else {
                final int maxminus2 = this.maxLocals() - 2;
                if (idx > maxminus2) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
                }
            }
        }
        
        @Override
        public void visitDSTORE(final DSTORE o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
            }
            else {
                final int maxminus2 = this.maxLocals() - 2;
                if (idx > maxminus2) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
                }
            }
        }
        
        @Override
        public void visitFieldInstruction(final FieldInstruction o) {
            try {
                this.indexValid(o, o.getIndex());
                final Constant c = this.constantPoolGen.getConstant(o.getIndex());
                if (!(c instanceof ConstantFieldref)) {
                    this.constraintViolated(o, "Indexing a constant that's not a CONSTANT_Fieldref but a '" + Pass3aVerifier.this.tostring(c) + "'.");
                }
                final String fieldName = o.getFieldName(this.constantPoolGen);
                final JavaClass jc = Repository.lookupClass(this.getObjectType(o).getClassName());
                Field[] fields = jc.getFields();
                Field f = null;
                for (final Field field : fields) {
                    if (field.getName().equals(fieldName)) {
                        final Type fType = Type.getType(field.getSignature());
                        final Type oType = o.getType(this.constantPoolGen);
                        if (fType.equals(oType)) {
                            f = field;
                            break;
                        }
                    }
                }
                if (f == null) {
                    final JavaClass[] superClasses;
                    final JavaClass[] superclasses = superClasses = jc.getSuperClasses();
                Label_0322:
                    for (final JavaClass superclass : superClasses) {
                        final Field[] fields2;
                        fields = (fields2 = superclass.getFields());
                        for (final Field field2 : fields2) {
                            if (field2.getName().equals(fieldName)) {
                                final Type fType2 = Type.getType(field2.getSignature());
                                final Type oType2 = o.getType(this.constantPoolGen);
                                if (fType2.equals(oType2)) {
                                    f = field2;
                                    if ((f.getAccessFlags() & 0x5) == 0x0) {
                                        f = null;
                                        break Label_0322;
                                    }
                                    break Label_0322;
                                }
                            }
                        }
                    }
                    if (f == null) {
                        this.constraintViolated(o, "Referenced field '" + fieldName + "' does not exist in class '" + jc.getClassName() + "'.");
                    }
                }
                else {
                    Type.getType(f.getSignature());
                    o.getType(this.constantPoolGen);
                }
            }
            catch (final ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }
        
        @Override
        public void visitFLOAD(final FLOAD o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitFSTORE(final FSTORE o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitGETSTATIC(final GETSTATIC o) {
            try {
                final String fieldName = o.getFieldName(this.constantPoolGen);
                final JavaClass jc = Repository.lookupClass(this.getObjectType(o).getClassName());
                final Field[] fields = jc.getFields();
                Field f = null;
                for (final Field field : fields) {
                    if (field.getName().equals(fieldName)) {
                        f = field;
                        break;
                    }
                }
                if (f == null) {
                    throw new AssertionViolatedException("Field '" + fieldName + "' not found in " + jc.getClassName());
                }
                if (!f.isStatic()) {
                    this.constraintViolated(o, "Referenced field '" + f + "' is not static which it should be.");
                }
            }
            catch (final ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }
        
        @Override
        public void visitIINC(final IINC o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitILOAD(final ILOAD o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitINSTANCEOF(final INSTANCEOF o) {
            this.indexValid(o, o.getIndex());
            final Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                this.constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
        }
        
        @Override
        public void visitINVOKEDYNAMIC(final INVOKEDYNAMIC o) {
            throw new UnsupportedOperationException("INVOKEDYNAMIC instruction is not supported at this time");
        }
        
        @Override
        public void visitInvokeInstruction(final InvokeInstruction o) {
            this.indexValid(o, o.getIndex());
            if (o instanceof INVOKEVIRTUAL || o instanceof INVOKESPECIAL || o instanceof INVOKESTATIC) {
                final Constant c = this.constantPoolGen.getConstant(o.getIndex());
                if (!(c instanceof ConstantMethodref)) {
                    this.constraintViolated(o, "Indexing a constant that's not a CONSTANT_Methodref but a '" + Pass3aVerifier.this.tostring(c) + "'.");
                }
                else {
                    final ConstantNameAndType cnat = (ConstantNameAndType)this.constantPoolGen.getConstant(((ConstantMethodref)c).getNameAndTypeIndex());
                    final ConstantUtf8 cutf8 = (ConstantUtf8)this.constantPoolGen.getConstant(cnat.getNameIndex());
                    if (cutf8.getBytes().equals("<init>") && !(o instanceof INVOKESPECIAL)) {
                        this.constraintViolated(o, "Only INVOKESPECIAL is allowed to invoke instance initialization methods.");
                    }
                    if (!cutf8.getBytes().equals("<init>") && cutf8.getBytes().startsWith("<")) {
                        this.constraintViolated(o, "No method with a name beginning with '<' other than the instance initialization methods may be called by the method invocation instructions.");
                    }
                }
            }
            else {
                final Constant c = this.constantPoolGen.getConstant(o.getIndex());
                if (!(c instanceof ConstantInterfaceMethodref)) {
                    this.constraintViolated(o, "Indexing a constant that's not a CONSTANT_InterfaceMethodref but a '" + Pass3aVerifier.this.tostring(c) + "'.");
                }
                final ConstantNameAndType cnat = (ConstantNameAndType)this.constantPoolGen.getConstant(((ConstantInterfaceMethodref)c).getNameAndTypeIndex());
                final String name = ((ConstantUtf8)this.constantPoolGen.getConstant(cnat.getNameIndex())).getBytes();
                if (name.equals("<init>")) {
                    this.constraintViolated(o, "Method to invoke must not be '<init>'.");
                }
                if (name.equals("<clinit>")) {
                    this.constraintViolated(o, "Method to invoke must not be '<clinit>'.");
                }
            }
            Type t = o.getReturnType(this.constantPoolGen);
            if (t instanceof ArrayType) {
                t = ((ArrayType)t).getBasicType();
            }
            if (t instanceof ObjectType) {
                final Verifier v = VerifierFactory.getVerifier(((ObjectType)t).getClassName());
                final VerificationResult vr = v.doPass2();
                if (vr.getStatus() != 1) {
                    this.constraintViolated(o, "Return type class/interface could not be verified successfully: '" + vr.getMessage() + "'.");
                }
            }
            final Type[] argumentTypes;
            final Type[] ts = argumentTypes = o.getArgumentTypes(this.constantPoolGen);
            for (int length = argumentTypes.length, i = 0; i < length; ++i) {
                final Type element = t = argumentTypes[i];
                if (t instanceof ArrayType) {
                    t = ((ArrayType)t).getBasicType();
                }
                if (t instanceof ObjectType) {
                    final Verifier v2 = VerifierFactory.getVerifier(((ObjectType)t).getClassName());
                    final VerificationResult vr2 = v2.doPass2();
                    if (vr2.getStatus() != 1) {
                        this.constraintViolated(o, "Argument type class/interface could not be verified successfully: '" + vr2.getMessage() + "'.");
                    }
                }
            }
        }
        
        @Override
        public void visitINVOKEINTERFACE(final INVOKEINTERFACE o) {
            try {
                final String className = o.getClassName(this.constantPoolGen);
                final JavaClass jc = Repository.lookupClass(className);
                final Method m = this.getMethodRecursive(jc, o);
                if (m == null) {
                    this.constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                }
                if (jc.isClass()) {
                    this.constraintViolated(o, "Referenced class '" + jc.getClassName() + "' is a class, but not an interface as expected.");
                }
            }
            catch (final ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }
        
        @Override
        public void visitINVOKESPECIAL(final INVOKESPECIAL o) {
            try {
                final String className = o.getClassName(this.constantPoolGen);
                final JavaClass jc = Repository.lookupClass(className);
                final Method m = this.getMethodRecursive(jc, o);
                if (m == null) {
                    this.constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                }
                JavaClass current = Repository.lookupClass(Pass3aVerifier.this.myOwner.getClassName());
                if (current.isSuper() && Repository.instanceOf(current, jc) && !current.equals(jc) && !o.getMethodName(this.constantPoolGen).equals("<init>")) {
                    int supidx = -1;
                    Method meth = null;
                    while (supidx != 0) {
                        supidx = current.getSuperclassNameIndex();
                        current = Repository.lookupClass(current.getSuperclassName());
                        final Method[] methods;
                        final Method[] meths = methods = current.getMethods();
                        for (final Method meth2 : methods) {
                            if (meth2.getName().equals(o.getMethodName(this.constantPoolGen)) && Type.getReturnType(meth2.getSignature()).equals(o.getReturnType(this.constantPoolGen)) && Arrays.equals(Type.getArgumentTypes(meth2.getSignature()), o.getArgumentTypes(this.constantPoolGen))) {
                                meth = meth2;
                                break;
                            }
                        }
                        if (meth != null) {
                            break;
                        }
                    }
                    if (meth == null) {
                        this.constraintViolated(o, "ACC_SUPER special lookup procedure not successful: method '" + o.getMethodName(this.constantPoolGen) + "' with proper signature not declared in superclass hierarchy.");
                    }
                }
            }
            catch (final ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }
        
        @Override
        public void visitINVOKESTATIC(final INVOKESTATIC o) {
            try {
                final String className = o.getClassName(this.constantPoolGen);
                final JavaClass jc = Repository.lookupClass(className);
                final Method m = this.getMethodRecursive(jc, o);
                if (m == null) {
                    this.constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                }
                else if (!m.isStatic()) {
                    this.constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' has ACC_STATIC unset.");
                }
            }
            catch (final ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }
        
        @Override
        public void visitINVOKEVIRTUAL(final INVOKEVIRTUAL o) {
            try {
                final String className = o.getClassName(this.constantPoolGen);
                final JavaClass jc = Repository.lookupClass(className);
                final Method m = this.getMethodRecursive(jc, o);
                if (m == null) {
                    this.constraintViolated(o, "Referenced method '" + o.getMethodName(this.constantPoolGen) + "' with expected signature '" + o.getSignature(this.constantPoolGen) + "' not found in class '" + jc.getClassName() + "'.");
                }
                if (!jc.isClass()) {
                    this.constraintViolated(o, "Referenced class '" + jc.getClassName() + "' is an interface, but not a class as expected.");
                }
            }
            catch (final ClassNotFoundException e) {
                Pass3aVerifier.this.addMessage("Unable to verify INVOKEVITUAL as cannot load target class: " + e.getCause());
            }
        }
        
        @Override
        public void visitISTORE(final ISTORE o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitLDC(final LDC ldc) {
            this.indexValid(ldc, ldc.getIndex());
            final Constant c = this.constantPoolGen.getConstant(ldc.getIndex());
            if (c instanceof ConstantClass) {
                Pass3aVerifier.this.addMessage("Operand of LDC or LDC_W is CONSTANT_Class '" + Pass3aVerifier.this.tostring(c) + "' - this is only supported in JDK 1.5 and higher.");
            }
            else if (!(c instanceof ConstantInteger) && !(c instanceof ConstantFloat) && !(c instanceof ConstantString)) {
                this.constraintViolated(ldc, "Operand of LDC or LDC_W must be one of CONSTANT_Integer, CONSTANT_Float or CONSTANT_String, but is '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
        }
        
        @Override
        public void visitLDC2_W(final LDC2_W o) {
            this.indexValid(o, o.getIndex());
            final Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantLong) && !(c instanceof ConstantDouble)) {
                this.constraintViolated(o, "Operand of LDC2_W must be CONSTANT_Long or CONSTANT_Double, but is '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
            try {
                this.indexValid(o, o.getIndex() + 1);
            }
            catch (final StaticCodeInstructionOperandConstraintException e) {
                throw new AssertionViolatedException("Does not BCEL handle that? LDC2_W operand has a problem.", e);
            }
        }
        
        @Override
        public void visitLLOAD(final LLOAD o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
            }
            else {
                final int maxminus2 = this.maxLocals() - 2;
                if (idx > maxminus2) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
                }
            }
        }
        
        @Override
        public void visitLoadClass(final LoadClass loadClass) {
            final ObjectType t = loadClass.getLoadClassType(this.constantPoolGen);
            if (t != null) {
                final Verifier v = VerifierFactory.getVerifier(t.getClassName());
                final VerificationResult vr = v.doPass1();
                if (vr.getStatus() != 1) {
                    this.constraintViolated((Instruction)loadClass, "Class '" + loadClass.getLoadClassType(this.constantPoolGen).getClassName() + "' is referenced, but cannot be loaded: '" + vr + "'.");
                }
            }
        }
        
        @Override
        public void visitLOOKUPSWITCH(final LOOKUPSWITCH o) {
            final int[] matchs = o.getMatchs();
            int max = Integer.MIN_VALUE;
            for (int i = 0; i < matchs.length; ++i) {
                if (matchs[i] == max && i != 0) {
                    this.constraintViolated(o, "Match '" + matchs[i] + "' occurs more than once.");
                }
                if (matchs[i] < max) {
                    this.constraintViolated(o, "Lookup table must be sorted but isn't.");
                }
                else {
                    max = matchs[i];
                }
            }
        }
        
        @Override
        public void visitLSTORE(final LSTORE o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative. [Constraint by JustIce as an analogon to the single-slot xLOAD/xSTORE instructions; may not happen anyway.]");
            }
            else {
                final int maxminus2 = this.maxLocals() - 2;
                if (idx > maxminus2) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-2 '" + maxminus2 + "'.");
                }
            }
        }
        
        @Override
        public void visitMULTIANEWARRAY(final MULTIANEWARRAY o) {
            this.indexValid(o, o.getIndex());
            final Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                this.constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
            final int dimensions2create = o.getDimensions();
            if (dimensions2create < 1) {
                this.constraintViolated(o, "Number of dimensions to create must be greater than zero.");
            }
            final Type t = o.getType(this.constantPoolGen);
            if (t instanceof ArrayType) {
                final int dimensions = ((ArrayType)t).getDimensions();
                if (dimensions < dimensions2create) {
                    this.constraintViolated(o, "Not allowed to create array with more dimensions ('" + dimensions2create + "') than the one referenced by the CONSTANT_Class '" + t + "'.");
                }
            }
            else {
                this.constraintViolated(o, "Expecting a CONSTANT_Class referencing an array type. [Constraint not found in The Java Virtual Machine Specification, Second Edition, 4.8.1]");
            }
        }
        
        @Override
        public void visitNEW(final NEW o) {
            this.indexValid(o, o.getIndex());
            final Constant c = this.constantPoolGen.getConstant(o.getIndex());
            if (!(c instanceof ConstantClass)) {
                this.constraintViolated(o, "Expecting a CONSTANT_Class operand, but found a '" + Pass3aVerifier.this.tostring(c) + "'.");
            }
            else {
                final ConstantUtf8 cutf8 = (ConstantUtf8)this.constantPoolGen.getConstant(((ConstantClass)c).getNameIndex());
                final Type t = Type.getType("L" + cutf8.getBytes() + ";");
                if (t instanceof ArrayType) {
                    this.constraintViolated(o, "NEW must not be used to create an array.");
                }
            }
        }
        
        @Override
        public void visitNEWARRAY(final NEWARRAY o) {
            final byte t = o.getTypecode();
            if (t != 4 && t != 5 && t != 6 && t != 7 && t != 8 && t != 9 && t != 10 && t != 11) {
                this.constraintViolated(o, "Illegal type code '" + Pass3aVerifier.this.tostring(t) + "' for 'atype' operand.");
            }
        }
        
        @Override
        public void visitPUTSTATIC(final PUTSTATIC o) {
            try {
                final String fieldName = o.getFieldName(this.constantPoolGen);
                final JavaClass jc = Repository.lookupClass(this.getObjectType(o).getClassName());
                final Field[] fields = jc.getFields();
                Field f = null;
                for (final Field field : fields) {
                    if (field.getName().equals(fieldName)) {
                        f = field;
                        break;
                    }
                }
                if (f == null) {
                    throw new AssertionViolatedException("Field '" + fieldName + "' not found in " + jc.getClassName());
                }
                if (f.isFinal() && !Pass3aVerifier.this.myOwner.getClassName().equals(this.getObjectType(o).getClassName())) {
                    this.constraintViolated(o, "Referenced field '" + f + "' is final and must therefore be declared in the current class '" + Pass3aVerifier.this.myOwner.getClassName() + "' which is not the case: it is declared in '" + o.getReferenceType(this.constantPoolGen) + "'.");
                }
                if (!f.isStatic()) {
                    this.constraintViolated(o, "Referenced field '" + f + "' is not static which it should be.");
                }
                final String methName = Repository.lookupClass(Pass3aVerifier.this.myOwner.getClassName()).getMethods()[Pass3aVerifier.this.methodNo].getName();
                if (!jc.isClass() && !methName.equals("<clinit>")) {
                    this.constraintViolated(o, "Interface field '" + f + "' must be set in a '" + "<clinit>" + "' method.");
                }
            }
            catch (final ClassNotFoundException e) {
                throw new AssertionViolatedException("Missing class: " + e, e);
            }
        }
        
        @Override
        public void visitRET(final RET o) {
            final int idx = o.getIndex();
            if (idx < 0) {
                this.constraintViolated(o, "Index '" + idx + "' must be non-negative.");
            }
            else {
                final int maxminus1 = this.maxLocals() - 1;
                if (idx > maxminus1) {
                    this.constraintViolated(o, "Index '" + idx + "' must not be greater than max_locals-1 '" + maxminus1 + "'.");
                }
            }
        }
        
        @Override
        public void visitTABLESWITCH(final TABLESWITCH o) {
        }
    }
}
